from typing import Optional

import requests

from ledgerwallet.hsmscript import HsmScript
from ledgerwallet.ledgerserver import LedgerServer
from ledgerwallet.proto.LedgerHSMServer_pb2 import Request, Response
from ledgerwallet.utils import serialize


class HsmServer(LedgerServer):
    def __init__(
        self, script: HsmScript, url="https://hsmprod.hardwarewallet.com/hsm/process"
    ):
        self.url = url
        self.script = script

        self.device_nonce: Optional[bytes] = None
        self.server_nonce: Optional[bytes] = None

        self.public_key: Optional[bytes] = None
        self.last_request_id: Optional[bytes] = None
        self.session = requests.Session()

    @staticmethod
    def _query_add_param(request: Request, alias: str, name: str, local: bool = False):
        param = request.remote_parameters.add()
        param.alias = alias
        param.name = name
        param.local = local

    def query(self, data=None, params=None) -> bytes:
        request = Request(
            reference=self.script.name, largeStack=self.script.use_large_stack
        )
        if self.last_request_id is not None:
            request.id = self.last_request_id.decode()

        if params:
            for param, value in params.items():
                self._query_add_param(request, param, value)
        elif self.script.default_params:
            for alias, name in self.script.default_params.items():
                self._query_add_param(request, alias, name)

        if data is not None:
            request.parameters = data

        req = self.session.post(self.url, request.SerializeToString())
        response = Response()
        # TODO: handle errors
        response.ParseFromString(req.content)

        self.last_request_id = response.id.encode() if response.id else b""
        if len(response.exception) != 0:
            raise Exception(f"HSM Error: {response.exception}")
        return response.response

    def send_nonce(self, nonce: bytes):
        assert len(nonce) == 8
        self.device_nonce = nonce

    def get_nonce(self) -> bytes:
        response = self.query()
        self.public_key, self.server_nonce = response[:65], response[65 : 65 + 8]

        return self.server_nonce

    def receive_certificate_chain(self):
        # Get signed public key
        response = self.query(self.device_nonce)
        server_signature = response
        public_key = self.public_key
        return [serialize(public_key) + serialize(server_signature)]

    def send_certificate_chain(self, chain):
        for certificate in chain:
            self.query(certificate)



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class App(_message.Message):
    __slots__ = ("flags", "hash", "hashCodeData", "name")
    FLAGS_FIELD_NUMBER: _ClassVar[int]
    HASH_FIELD_NUMBER: _ClassVar[int]
    HASHCODEDATA_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    flags: int
    hash: bytes
    hashCodeData: bytes
    name: str
    def __init__(self, flags: _Optional[int] = ..., hash: _Optional[bytes] = ..., hashCodeData: _Optional[bytes] = ..., name: _Optional[str] = ...) -> None: ...

class AppList(_message.Message):
    __slots__ = ("list",)
    LIST_FIELD_NUMBER: _ClassVar[int]
    list: _containers.RepeatedCompositeFieldContainer[App]
    def __init__(self, list: _Optional[_Iterable[_Union[App, _Mapping]]] = ...) -> None: ...



import os
from typing import Optional

from ledgerwallet.crypto.ecc import PrivateKey, PublicKey
from ledgerwallet.ledgerserver import LedgerServer
from ledgerwallet.utils import serialize, unserialize

CERT_ROLE_SIGNER = 1
CERT_ROLE_DEVICE = 2
CERT_ROLE_SIGNER_EPHEMERAL = 0x11
CERT_ROLE_DEVICE_EPHEMERAL = 0x12


class SimpleServer(LedgerServer):
    def __init__(self, master_private: PrivateKey, cert_chain=None):
        self.device_nonce: Optional[bytes] = None
        self.server_nonce: Optional[bytes] = None
        self.master_private = master_private
        self.master_public = master_private.pubkey.serialize(False)
        self.shared_secret: Optional[bytes] = None
        self.ephemeral_private: Optional[bytes] = None

        self.cert_chain = cert_chain

    def receive_certificate_chain(self):
        cert_chain = []
        if self.cert_chain is not None:
            cert_chain.append(self.cert_chain)
        else:
            data_to_sign = bytes([CERT_ROLE_SIGNER]) + self.master_public
            master_signature = self.master_private.sign(data_to_sign)
            cert_chain.append(
                serialize(self.master_public) + serialize(master_signature)
            )

        # Provide the ephemeral certificate, signed with the master public key
        self.ephemeral_private = PrivateKey()
        ephemeral_public = self.ephemeral_private.pubkey.serialize(compressed=False)
        # print("Using ephemeral key {}".format(ephemeral_public.hex()))

        data_to_sign = (
            bytes([CERT_ROLE_SIGNER_EPHEMERAL])
            + self.server_nonce
            + self.device_nonce
            + ephemeral_public
        )
        signature = self.master_private.sign(data_to_sign)
        cert_chain.append(serialize(ephemeral_public) + serialize(signature))

        return cert_chain

    def send_certificate_chain(self, chain):
        assert len(chain) == 2

        last_dev_pub_key = PublicKey(self.master_public)
        for i, item in enumerate(chain):
            certificate_header, item = unserialize(item)
            certificate_public_key, item = unserialize(item)
            certificate_signature_array, _ = unserialize(item)
            certificate_signature = certificate_signature_array

            # First cert contains a header field which holds the certificate's
            # public key role
            if i == 0:
                # device_public_key = certificate_public_key
                certificate_signed_data = (
                    bytes([CERT_ROLE_DEVICE])
                    + certificate_header
                    + certificate_public_key
                )
            # Could check if the device certificate is signed by the issuer public key
            # ephemeral key certificate
            else:
                certificate_signed_data = (
                    bytes([CERT_ROLE_DEVICE_EPHEMERAL])
                    + self.device_nonce
                    + self.server_nonce
                    + certificate_public_key
                )

            if not last_dev_pub_key.verify(
                certificate_signed_data, certificate_signature
            ):
                """
                if index == 0:
                    # Not an error if loading from user key
                    print("Broken certificate chain - loading from user key")
                else:
                    raise Exception("Broken certificate chain")
                """
                if i != 0:
                    raise Exception("Broken certificate chain")

            last_dev_pub_key = PublicKey(certificate_public_key)
        self.shared_secret = self.ephemeral_private.exchange(last_dev_pub_key)

    def get_nonce(self) -> bytes:
        self.server_nonce = os.urandom(8)
        return self.server_nonce

    def send_nonce(self, nonce: bytes):
        assert len(nonce) == 8
        self.device_nonce = nonce

    def get_shared_secret(self):
        return self.shared_secret



from unittest import TestCase

from ledgerwallet.crypto import ecc

RAW_PRIVATE = bytes.fromhex(
    "c2cdf0a8b0a83b35ace53f097b5e6e6a0a1f2d40535eff1cf434f52a43d59d8f"
)
RAW_PUBLIC = bytes.fromhex(
    "6fcc37ea5e9e09fec6c83e5fbd7a745e3eee81d16ebd861c9e66f55518c19798"
    + "4e9f113c07f875691df8afc1029496fc4cb9509b39dcd38f251a83359cc8b4f7"
)


class PrivateKeyTest(TestCase):
    def setUp(self):
        self.key = ecc.PrivateKey(RAW_PRIVATE)

    def test_pubkey(self):
        self.assertEqual(
            self.key.pubkey.serialize(compressed=False), bytes([0x04]) + RAW_PUBLIC
        )

    def test_serialize(self):
        self.assertEqual(self.key.serialize(), RAW_PRIVATE)

    def test_sign(self):
        blob = b"someblobofdata"
        signature = self.key.sign(blob)
        self.assertTrue(self.key.pubkey.verify(blob, signature))


class PublickKeyTest(TestCase):
    def setUp(self):
        self.public = bytes([0x04]) + RAW_PUBLIC
        self.key = ecc.PublicKey(self.public)

    def test___init__fail(self):
        with self.assertRaises(ValueError):
            ecc.PublicKey(b"")  # not 65-bytes long
        with self.assertRaises(ValueError):
            self.public = bytes([0x00]) + self.public[1:]
            ecc.PublicKey(self.public)  # must start with 0x04

    def test_serialize_uncompressed(self):
        self.assertEqual(self.key.serialize(compressed=False), self.public)

    def test_serialize_compressed(self):
        with self.assertRaises(NotImplementedError):
            self.key.serialize()



import os
import socket

from .device import Device


class TcpDevice(Device):
    LEDGER_PROXY_ADDRESS = "127.0.0.1"
    LEDGER_PROXY_PORT = 1237

    def __init__(self, path: str):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server, port = path.split(":")
        self.server = server
        self.port = int(port)

    @classmethod
    def enumerate_devices(cls):
        if "LEDGER_PROXY_ADDRESS" in os.environ and "LEDGER_PROXY_PORT" in os.environ:
            return [
                TcpDevice(
                    "{0:s}:{1:d}".format(
                        os.environ["LEDGER_PROXY_ADDRESS"],
                        int(os.environ["LEDGER_PROXY_PORT"]),
                    )
                )
            ]
        else:
            return []

    def open(self):
        self.socket.connect((self.server, self.port))

    def write(self, data: bytes):
        # data is prefixed by its size
        data_to_send = int.to_bytes(len(data), 4, "big") + data
        self.socket.send(data_to_send)

    def read(self, timeout: int = 0) -> bytes:
        packet_len = int.from_bytes(self.socket.recv(4), "big")
        return self.socket.recv(packet_len + 2)

    def exchange(self, data: bytes, timeout: int = 0):
        self.write(data)
        return self.read()

    def close(self):
        self.socket.close()



import sys

from ..utils import LedgerIns, VersionInfo
from .device import Device


class FileDevice(Device):
    def __init__(self, target_id, out=None):
        if out is None:
            out = sys.stdout
        t_id = int(target_id, 16)
        self.version_info = VersionInfo.build(
            dict(target_id=t_id, se_version="0", flags=0, mcu_version="0")
        )
        self.buffer = None
        self.out = out

    @classmethod
    def enumerate_devices(cls):
        return None

    def open(self):
        pass

    def write(self, data: bytes):
        self.buffer = data
        if not self.buffer[1] == LedgerIns.GET_VERSION:
            print(data.hex(), file=self.out)

    def read(self, timeout: int = 0) -> bytes:
        if self.buffer[1] == LedgerIns.GET_VERSION:
            return self.version_info + b"\x90\x00"
        return b"\x00\x00\x00\x02\x90\x00"

    def exchange(self, data: bytes, timeout: int = 0) -> bytes:
        self.write(data)
        return self.read()

    def close(self):
        if self.out:
            self.out.close()



from unittest import TestCase

from construct import StreamError

from ledgerwallet.params import (
    Asn1Length,
    Bip32Path,
    Dependencies,
    Dependency,
    DerivationPath,
)


class Asn1LengthTest(TestCase):
    sample = {
        0: bytes.fromhex("00"),
        4: bytes.fromhex("04"),
        127: bytes.fromhex("7f"),
        128: bytes.fromhex("8180"),
        160: bytes.fromhex("81a0"),
        255: bytes.fromhex("81ff"),
        256: bytes.fromhex("820100"),
    }

    def test_parse(self):
        for k, v in self.sample.items():
            self.assertEqual(Asn1Length.parse(v), k)

    def test_build(self):
        for k, v in self.sample.items():
            self.assertEqual(Asn1Length.build(k), v)


class Bip32PathTest(TestCase):
    sample = {
        "1": bytes.fromhex("01 00000001"),
        "1'": bytes.fromhex("01 80000001"),
        "0'/0": bytes.fromhex("02 80000000 00000000"),
        "44'/91223'/2": bytes.fromhex("03 8000002c 80016457 00000002"),
        "44'/0'/0'/1/400": bytes.fromhex(
            "05 8000002c 80000000 80000000 00000001 00000190"
        ),
    }

    def test_parse(self):
        for k, v in self.sample.items():
            self.assertEqual(Bip32Path.parse(v), k)

    def test_parse_empty(self):
        # not in sample as not the parse/build behavior is not symmetrical
        self.assertEqual(Bip32Path.parse(bytes.fromhex("00")), str())

    def test_build(self):
        for k, v in self.sample.items():
            self.assertEqual(Bip32Path.build(k), v)

    def test_parse_error(self):
        errors = [
            b"",  # empty string is not parsable
            bytes.fromhex("01"),  # expecting 4 more bytes (Int32ub)
        ]
        for error in errors:
            with self.assertRaises(StreamError):
                Bip32Path.parse(error)


class DerivationPathTest(TestCase):
    def test_parse_empty(self):
        result = DerivationPath.parse(bytes.fromhex("01 00"))
        self.assertFalse(result.curve.secp256k1)
        self.assertFalse(result.curve.prime256r1)
        self.assertFalse(result.curve.ed25519)
        self.assertFalse(result.curve.bls12381g1)
        self.assertFalse(result.paths)

    def test_parse_keys(self):
        # [secp256k1, prime256r1, ed25519, bls12381g1]
        # fmt: off
        keys = [
            [False, False, False, False],
            [True,  False, False, False],
            [False, True,  False, False],
            [True,  True,  False, False],
            [False, False, True,  False],
            [True,  False, True,  False],
            [False, True,  True,  False],
            [True,  True,  True,  False],
            [False, False, False, True],
            [True,  False, False, True],
            [False, True,  False, True],
            [True,  True,  False, True],
            [False, False, True,  True],
            [True,  False, True,  True],
            [False, True,  True,  True],
            [True,  True,  True,  True],
        ]
        # fmt: on
        for i in range(8):
            result = DerivationPath.parse(bytes([1, i]))
            self.assertListEqual(
                [
                    result.curve.secp256k1,
                    result.curve.prime256r1,
                    result.curve.ed25519,
                    result.curve.bls12381g1,
                ],
                keys[i],
            )
        # bls12381g1 store on 5th bit (16), not 4th (8)
        # so there is a bit gap from 8 to 16
        for i in range(16, 24):
            result = DerivationPath.parse(bytes([1, i]))
            self.assertListEqual(
                [
                    result.curve.secp256k1,
                    result.curve.prime256r1,
                    result.curve.ed25519,
                    result.curve.bls12381g1,
                ],
                keys[i - 8],
            )

    def test_parse_with_paths(self):
        path1 = (
            bytes.fromhex("05 8000002c 80000000 80000000 00000001 00000190"),
            "44'/0'/0'/1/400",
        )
        path2 = (bytes.fromhex("03 8000002c 80000000 000000ff"), "44'/0'/255")
        key = bytes.fromhex("11")
        size = len(path1[0]) + len(path2[0]) + len(key)
        result = DerivationPath.parse(bytes([size]) + key + path1[0] + path2[0])
        self.assertTrue(result.curve.secp256k1)
        self.assertFalse(result.curve.prime256r1)
        self.assertFalse(result.curve.ed25519)
        self.assertTrue(result.curve.bls12381g1)
        self.assertEqual(result.paths, [path1[1], path2[1]])

    def test_parse_error(self):
        errors = [
            b"",  # empty string is not parsable
            bytes.fromhex("01"),  # expecting more bytes (curve)
        ]
        for error in errors:
            with self.assertRaises(StreamError):
                DerivationPath.parse(error)


class DependencyTest(TestCase):
    def test_parse_no_version(self):
        name = "name"
        asn1_name = bytes([len(name)]) + name.encode()
        result = Dependency.parse(bytes([len(asn1_name)]) + asn1_name)
        self.assertEqual(result.name, name)
        self.assertIsNone(result.version)

    def test_parse_with_version(self):
        name = "name"
        version = "1.0.1"
        asn1_name = bytes([len(name)]) + name.encode()
        asn1_version = bytes([len(version)]) + version.encode()
        result = Dependency.parse(
            bytes([len(asn1_name + asn1_version)]) + asn1_name + asn1_version
        )
        self.assertEqual(result.name, name)
        self.assertEqual(result.version, version)


class DependenciesTest(TestCase):
    def test_parse(self):
        name1 = "name1"
        version = "1.0.1"
        asn1_name1 = bytes([len(name1)]) + name1.encode()
        asn1_version = bytes([len(version)]) + version.encode()
        dep1 = bytes([len(asn1_name1 + asn1_version)]) + asn1_name1 + asn1_version
        name2 = "name2"
        asn1_name2 = bytes([len(name2)]) + name2.encode()
        dep2 = bytes([len(asn1_name2)]) + asn1_name2
        result = Dependencies.parse(bytes([len(dep1 + dep2)]) + dep1 + dep2)
        self.assertEqual(result[0].name, name1)
        self.assertEqual(result[0].version, version)
        self.assertEqual(result[1].name, name2)
        self.assertIsNone(result[1].version)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: ledgerwallet/proto/LedgerHSMServer.proto
# Protobuf Python Version: 5.28.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    1,
    '',
    'ledgerwallet/proto/LedgerHSMServer.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n(ledgerwallet/proto/LedgerHSMServer.proto\x12\rbluehsmserver\"7\n\tParameter\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05\x61lias\x18\x02 \x01(\t\x12\r\n\x05local\x18\x03 \x01(\x08\"\xa1\x01\n\x07Request\x12\n\n\x02id\x18\x01 \x01(\t\x12\x12\n\nparameters\x18\x02 \x01(\x0c\x12\x11\n\treference\x18\x03 \x01(\t\x12\x0b\n\x03\x65lf\x18\x04 \x01(\x0c\x12\r\n\x05\x63lose\x18\x05 \x01(\x08\x12\x12\n\nlargeStack\x18\x06 \x01(\x08\x12\x33\n\x11remote_parameters\x18\x07 \x03(\x0b\x32\x18.bluehsmserver.Parameter\"L\n\x08Response\x12\n\n\x02id\x18\x01 \x01(\t\x12\x10\n\x08response\x18\x02 \x01(\x0c\x12\x0f\n\x07message\x18\x03 \x01(\t\x12\x11\n\texception\x18\x04 \x01(\tb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ledgerwallet.proto.LedgerHSMServer_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_PARAMETER']._serialized_start=59
  _globals['_PARAMETER']._serialized_end=114
  _globals['_REQUEST']._serialized_start=117
  _globals['_REQUEST']._serialized_end=278
  _globals['_RESPONSE']._serialized_start=280
  _globals['_RESPONSE']._serialized_end=356
# @@protoc_insertion_point(module_scope)



import hashlib
from typing import Optional

import ecdsa.ellipticcurve
import ecdsa.util
from ecdsa.curves import SECP256k1
from ecdsa.keys import SigningKey, VerifyingKey


class PublicKey(object):
    def __init__(self, pubkey: bytes):
        if len(pubkey) != 64 + 1 or pubkey[0] != 0x04:
            raise ValueError

        self.vk = VerifyingKey.from_string(pubkey[1:], SECP256k1, validate_point=True)

    def serialize(self, compressed: bool = True) -> bytes:
        if compressed:
            raise NotImplementedError
        return b"\x04" + self.vk.to_string()

    def verify(
        self, msg: bytes, raw_sig: bytes, raw=False, hashfunc=hashlib.sha256
    ) -> bool:
        try:
            if not raw:
                return self.vk.verify(raw_sig, msg, hashfunc, ecdsa.util.sigdecode_der)
            else:
                return self.vk.verify_digest(raw_sig, msg, ecdsa.util.sigdecode_der)
        except ecdsa.keys.BadSignatureError:
            return False


class PrivateKey(object):
    def __init__(self, sk: Optional[bytes] = None):
        if sk is None:
            self.sk = SigningKey.generate(SECP256k1)
        else:
            self.sk = SigningKey.from_string(sk, SECP256k1)

    @property
    def pubkey(self) -> PublicKey:
        return PublicKey(b"\x04" + self.sk.get_verifying_key().to_string())

    def serialize(self) -> bytes:
        return self.sk.to_string()

    def sign(self, msg, raw=False, hashfunc=hashlib.sha256) -> bytes:
        if not raw:
            signature = self.sk.sign(
                msg, hashfunc=hashfunc, sigencode=ecdsa.util.sigencode_der_canonize
            )
        else:
            signature = self.sk.sign_digest(
                msg, sigencode=ecdsa.util.sigencode_der_canonize
            )
        return signature

    def exchange(self, public_key: PublicKey) -> bytes:
        # ECDH as computed by libsecpk256k1
        point = self.sk.privkey.secret_multiplier * public_key.vk.pubkey.point
        if point.y() & 1 == 1:
            msg = b"\x03" + ecdsa.util.number_to_string(point.x(), SECP256k1.order)
        else:
            msg = b"\x02" + ecdsa.util.number_to_string(point.x(), SECP256k1.order)
        md = hashlib.sha256(msg)
        return md.digest()



import struct

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import constant_time, hashes, serialization
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

BLOCK_SIZE = 16


def iso9797_pad(data: bytes) -> bytes:
    padding_len = BLOCK_SIZE - len(data) % BLOCK_SIZE
    return data + b"\x80" + b"\x00" * (padding_len - 1)


def iso9797_unpad(data: bytes) -> bytes:
    assert len(data) >= BLOCK_SIZE and len(data) % BLOCK_SIZE == 0

    last_block = data[-BLOCK_SIZE:]
    while len(last_block) > 0:
        if last_block[-1] == 0x80:
            return data[:-BLOCK_SIZE] + last_block[:-1]
        elif last_block[-1] != 0:
            raise ValueError("Invalid padding")
        last_block = last_block[:-1]
    raise ValueError("Invalid padding")


class SCP(object):
    SCP_MAC_LENGTH = 14
    VERSION = 3

    def __init__(self, secret: bytes):
        self.enc_key = self._derive_key(secret, 0, 16)
        self.mac_key = self._derive_key(secret, 1, 16)
        self.enc_iv = b"\x00" * BLOCK_SIZE
        self.mac_iv = b"\x00" * BLOCK_SIZE

    @staticmethod
    def _derive_key(secret: bytes, index: int, key_len: int) -> bytes:
        retry = 0

        # secp256k1 curve order
        curve_order = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEBAAEDCE6AF48A03BBFD25E8CD0364141
        while True:
            digest = hashes.Hash(hashes.SHA256(), backend=default_backend())
            digest.update(struct.pack(">IB", index, retry))
            digest.update(secret)
            md = digest.finalize()

            private_value = int.from_bytes(md, "big")
            if private_value < curve_order and private_value != 0:
                break
            retry += 1

        key = ec.derive_private_key(private_value, ec.SECP256K1(), default_backend())
        public_key = key.public_key()

        digest = hashes.Hash(hashes.SHA256(), backend=default_backend())
        digest.update(
            public_key.public_bytes(
                serialization.Encoding.X962,
                serialization.PublicFormat.UncompressedPoint,
            )
        )
        return digest.finalize()[:key_len]

    def _decrypt_data(self, data: bytes) -> bytes:
        cipher = Cipher(
            algorithms.AES(self.enc_key), modes.CBC(self.enc_iv), default_backend()
        )
        self.enc_iv = data[-BLOCK_SIZE:]
        decrypted_data = cipher.decryptor().update(data)
        return decrypted_data

    def _encrypt_data(self, data: bytes) -> bytes:
        cipher = Cipher(
            algorithms.AES(self.enc_key), modes.CBC(self.enc_iv), default_backend()
        )
        encrypted_data = cipher.encryptor().update(data)
        self.enc_iv = encrypted_data[-16:]
        return encrypted_data

    def _compute_cbc_mac(self, data: bytes) -> bytes:
        cipher = Cipher(
            algorithms.AES(self.mac_key), modes.CBC(self.mac_iv), default_backend()
        )
        encrypted_data = cipher.encryptor().update(data)
        self.mac_iv = encrypted_data[-BLOCK_SIZE:]
        return encrypted_data[-BLOCK_SIZE:]

    def _verify_cbc_mac(self, data: bytes, mac: bytes) -> bool:
        computed_mac = self._compute_cbc_mac(data)
        computed_mac = computed_mac[-len(mac) :]
        return constant_time.bytes_eq(computed_mac, mac)

    def wrap(self, data: bytes) -> bytes:
        padded_data = iso9797_pad(data)
        encrypted_data = self._encrypt_data(padded_data)
        mac_data = self._compute_cbc_mac(encrypted_data)
        encrypted_data += mac_data[
            -self.SCP_MAC_LENGTH :
        ]  # only append part of the mac
        return encrypted_data

    def unwrap(self, data: bytes) -> bytes:
        if len(data) == 0:
            return b""

        encrypted_data, mac = data[: -self.SCP_MAC_LENGTH], data[-self.SCP_MAC_LENGTH :]
        if not self._verify_cbc_mac(encrypted_data, mac):
            raise Exception("Invalid SCP MAC")
        data = self._decrypt_data(encrypted_data)
        return iso9797_unpad(data)


class FakeSCP:
    def __init__(self):
        pass

    @staticmethod
    def identity_wrap(data: bytes) -> bytes:
        return data

    def wrap(self, data):
        return self.identity_wrap(data)

    def unwrap(self, data):
        return self.identity_wrap(data)



from unittest import TestCase

from ledgerwallet import utils


class UtilsModuleTest(TestCase):
    def test_serialize(self):
        sample = b"some bytes"
        result = utils.serialize(sample)
        self.assertEqual(result[0], len(result) - 1)
        self.assertEqual(result[0], len(sample))

    def test_unserialize(self):
        sample = bytes.fromhex("0304050607")
        result, rest = utils.unserialize(sample)
        self.assertEqual(result, sample[1 : 1 + sample[0]])
        self.assertEqual(rest, sample[1 + sample[0] :])

    def test_unserialize_too_small(self):
        sample = bytes.fromhex("01")
        with self.assertRaises(AssertionError):
            utils.unserialize(sample)

    def test_flags_to_string(self):
        self.assertEqual(
            utils.flags_to_string(5333),
            "issuer,signed,derive_master,global_pin,debug,custom_ca,no_run",
        )

    def test_version_info_extended_build_and_parse(self):
        # Build an extended VersionInfo with optional fields present
        payload = utils.VersionInfo.build(
            dict(
                target_id=0x33000004,  # NANO_X
                se_version="1.2.3",
                flags=0,
                mcu_version="2.3.4",
                mcu_bl_version="5.6.7",
                hw_version="\x01",  # 01.00
                language="\x00",  # en
                _recover_state_len=0x01,
                recover_state=0x02,
            )
        )

        parsed = utils.VersionInfo.parse(payload)
        self.assertEqual(int(parsed.target_id), 0x33000004)
        self.assertEqual(parsed.se_version, "1.2.3")
        self.assertEqual(parsed.mcu_version, "2.3.4")
        self.assertEqual(parsed.mcu_bl_version, "5.6.7")
        self.assertEqual(parsed.hw_version, "\x01")
        self.assertEqual(parsed.language, "\x00")
        self.assertEqual(parsed._recover_state_len, 0x01)
        self.assertEqual(parsed.recover_state, 0x02)

    def test_version_info_minimal_build_and_parse(self):
        # Build a minimal VersionInfo with only required fields
        payload = utils.VersionInfo.build(
            dict(
                target_id=0x31100004,
                se_version="0",
                flags=0,
                mcu_version="0",
            )
        )

        parsed = utils.VersionInfo.parse(payload)
        self.assertEqual(int(parsed.target_id), 0x31100004)
        self.assertEqual(parsed.se_version, "0")
        self.assertEqual(parsed.mcu_version, "0")
        self.assertIsNone(parsed.mcu_bl_version)
        # When target_id is not Nano X, hw_version is an empty bytes Const
        self.assertEqual(parsed.hw_version, b"")
        self.assertIsNone(parsed.language)
        self.assertIsNone(parsed._recover_state_len)
        self.assertIsNone(parsed.recover_state)



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Parameter(_message.Message):
    __slots__ = ("name", "alias", "local")
    NAME_FIELD_NUMBER: _ClassVar[int]
    ALIAS_FIELD_NUMBER: _ClassVar[int]
    LOCAL_FIELD_NUMBER: _ClassVar[int]
    name: str
    alias: str
    local: bool
    def __init__(self, name: _Optional[str] = ..., alias: _Optional[str] = ..., local: bool = ...) -> None: ...

class Request(_message.Message):
    __slots__ = ("id", "parameters", "reference", "elf", "close", "largeStack", "remote_parameters")
    ID_FIELD_NUMBER: _ClassVar[int]
    PARAMETERS_FIELD_NUMBER: _ClassVar[int]
    REFERENCE_FIELD_NUMBER: _ClassVar[int]
    ELF_FIELD_NUMBER: _ClassVar[int]
    CLOSE_FIELD_NUMBER: _ClassVar[int]
    LARGESTACK_FIELD_NUMBER: _ClassVar[int]
    REMOTE_PARAMETERS_FIELD_NUMBER: _ClassVar[int]
    id: str
    parameters: bytes
    reference: str
    elf: bytes
    close: bool
    largeStack: bool
    remote_parameters: _containers.RepeatedCompositeFieldContainer[Parameter]
    def __init__(self, id: _Optional[str] = ..., parameters: _Optional[bytes] = ..., reference: _Optional[str] = ..., elf: _Optional[bytes] = ..., close: bool = ..., largeStack: bool = ..., remote_parameters: _Optional[_Iterable[_Union[Parameter, _Mapping]]] = ...) -> None: ...

class Response(_message.Message):
    __slots__ = ("id", "response", "message", "exception")
    ID_FIELD_NUMBER: _ClassVar[int]
    RESPONSE_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    EXCEPTION_FIELD_NUMBER: _ClassVar[int]
    id: str
    response: bytes
    message: str
    exception: str
    def __init__(self, id: _Optional[str] = ..., response: _Optional[bytes] = ..., message: _Optional[str] = ..., exception: _Optional[str] = ...) -> None: ...



import sys
from pathlib import Path
from unittest import TestCase
from unittest.mock import patch

if sys.version_info >= (3, 11):
    import tomllib
else:
    import toml as tomllib

from ledgerwallet.manifest_toml import AppManifestToml


class ManifestTestToml(TestCase):
    def setUp(self):
        self.data_dir = Path(__file__).parent / "data"
        self.toml_path = self.data_dir / "manifest.toml"
        self.assertTrue(self.toml_path.is_file())
        self.toml_manifest = AppManifestToml(str(self.toml_path))

    def test___init__(self):
        self.assertEqual(self.toml_manifest.path, str(self.toml_path.parent))

        if sys.version_info >= (3, 11):
            with self.toml_path.open("rb") as filee:
                self.assertEqual(self.toml_manifest.dic, tomllib.load(filee))
        else:
            with self.toml_path.open("r") as filee:
                self.assertEqual(self.toml_manifest.dic, tomllib.load(filee))

    def test___init__error(self):
        with self.assertRaises(FileNotFoundError):
            AppManifestToml("/this/path/should/not/exist")

    def test_app_name(self):
        self.assertEqual(self.toml_manifest.app_name, "some name")

    def test_data_size(self):
        self.assertEqual(self.toml_manifest.data_size("1234"), 42)

    def test_get_application_flags(self):
        self.assertEqual(self.toml_manifest.get_application_flags("1234"), 0x9999)

    def test_get_binary(self):
        self.assertEqual(
            self.toml_manifest.get_binary("1234"), str(self.data_dir / "some binary")
        )

    def test_assert_compatible_device(self):
        self.toml_manifest.assert_compatible_device(1234)

    def test_properties_with_minimal_manifest(self):
        manifest_toml = AppManifestToml(str(self.data_dir / "minimal_manifest.toml"))
        self.assertEqual(manifest_toml.app_name, "")
        self.assertEqual(manifest_toml.data_size("1234"), 0)
        self.assertEqual(manifest_toml.get_application_flags("1234"), 0)
        self.assertEqual(
            manifest_toml.get_binary("1234"), str(self.data_dir / "some binary")
        )
        self.assertEqual(manifest_toml.serialize_parameters("1234"), b"")

    def test_serialize_parameters(self):
        # fmt: off
        expected = bytes.fromhex(
            "01" +  # BolosTag 'AppName'
            "09" + "736f6d65206e616d65" +  # "some name"
            "02" +  # BolosTag 'Version'
            "0b" + "322e392e342d6465627567" +  # 2.9.4-debug
            "03" +  # BolosTag 'Icon'
            "04" + "01020304" +  # mocked icon
            "04" +  # BolosTag 'DerivationPath'
            "23" +  # 35: following size
            "05" +  # secp256k1 (1) + ed25519 (4)
            "03" + "8000002c80000000000000ff" +                # "44'/0'/255"
            "05" + "8000002c80000000800000000000000100000190"  # "44'/0'/0'/1/400"
        )
        # fmt: on
        with patch(
            "ledgerwallet.manifest_toml.icon_from_file",
            lambda x, y, z: b"\x01\x02\x03\x04",
        ):
            result_toml = self.toml_manifest.serialize_parameters("1234")
        self.assertEqual(result_toml, expected)



import enum

from construct import (
    Adapter,
    Byte,
    Construct,
    Enum,
    FlagsEnum,
    GreedyBytes,
    GreedyRange,
    Int32ub,
    IntegerError,
    Optional,
    PascalString,
    Prefixed,
    PrefixedArray,
    Struct,
    Switch,
)
from construct.core import (
    byte2int,
    singleton,
    stream_read,
    stream_write,
    swapbytes,
    this,
)
from construct.lib import int2byte, integertypes


# noinspection PyAbstractClass
@singleton
class Asn1Length(Construct):
    def _parse(self, stream, context, path):
        byte = byte2int(stream_read(stream, 1, path))
        if byte & 0x80 == 0:
            return byte

        num_bytes = byte & ~0x80
        encoded_len = stream_read(stream, num_bytes, path)
        num = 0
        for len_byte in encoded_len:
            num = (num << 8) + len_byte
        return num

    def _build(self, obj, stream, context, path):
        if not isinstance(obj, integertypes):
            raise IntegerError("value is not an integer")
        if obj < 0:
            raise IntegerError(
                "asn1length cannot build from negative number: %r" % (obj,)
            )
        num = obj
        if num < 0x80:
            stream_write(stream, int2byte(num), 1, path)
        else:
            acc = b""
            while num != 0:
                acc += int2byte(num & 0xFF)
                num >>= 8
            stream_write(stream, int2byte(0x80 | len(acc)), 1, path)
            stream_write(stream, swapbytes(acc), len(acc), path)
        return obj

    def _emitprimitivetype(self, ksy, bitwise):
        return "asn1_der_len"


# A byte with value < 0x80
@singleton
class LowByte(Construct):
    def _parse(self, stream, context, path):
        b = byte2int(stream_read(stream, 1, path))
        if b >= 0x80:
            raise IntegerError("BIP32 path length prefix must be < 0x80")
        return b

    def _build(self, obj, stream, context, path):
        if not isinstance(obj, int):
            raise IntegerError("length must be an integer")
        if not (0 <= obj < 0x80):
            raise IntegerError("BIP32 path length must be < 0x80")
        stream_write(stream, int2byte(obj), 1, path)
        return obj


# A byte with value >= 0x80, decoded as (byte - 0x80)
# This is used to encode the length of SLIP-21 paths
@singleton
class Slip21LenByte(Construct):
    def _parse(self, stream, context, path):
        b = byte2int(stream_read(stream, 1, path))
        if b < 0x80:
            raise IntegerError("SLIP-21 length prefix must be >= 0x80")
        return b - 0x80

    def _build(self, obj, stream, context, path):
        if not isinstance(obj, int):
            raise IntegerError("length must be an integer")
        if not (0 <= obj <= 0x7F):
            raise IntegerError("SLIP-21 decoded length must be in [0, 0x7F]")
        stream_write(stream, int2byte(0x80 + obj), 1, path)
        return obj


class Slip21PathAdapter(Adapter):
    def _decode(self, obj, context, path):
        # obj is now a list/array of bytes (ints 0–255)
        if not obj:
            return str()

        # First byte must be zero prefix
        if obj[0] != 0x00:
            raise IntegerError("invalid SLIP-21 path prefix")

        # Remaining bytes are UTF-8 chars
        return bytes(obj[1:]).decode("utf-8")

    def _encode(self, obj, context, path):
        if not isinstance(obj, str):
            raise IntegerError("SLIP-21 path must be a string")

        payload = b"\0" + obj.encode("utf-8")
        length = len(payload)
        if length > 0x7F:
            raise IntegerError("SLIP-21 path too long")

        # Return list of ints so PrefixedArray(Slip21LenByte, Byte) can handle it
        return list(payload)


# noinspection PyAbstractClass
class Bip32PathAdapter(Adapter):
    def _decode(self, obj, context, path):
        out = list()
        for element in obj:
            if element & 0x80000000:
                out.append(str(element & 0x7FFFFFFF) + "'")
            else:
                out.append(str(element))
        return "/".join(out)

    def _encode(self, obj, context, path):
        out = list()
        elements = obj.split("/")
        if elements[0] == "m":
            elements = elements[1:]
        for element in elements:
            if element.endswith("'"):
                out.append(0x80000000 | int(element[:-1]))
            else:
                out.append(int(element))
        return out


Bip32Path = Bip32PathAdapter(PrefixedArray(LowByte, Int32ub))
Slip21Path = Slip21PathAdapter(PrefixedArray(Slip21LenByte, Byte))

PrefixedString = PascalString(Asn1Length, "utf8")

AppName = PrefixedString
Version = PrefixedString
Icon = Prefixed(Asn1Length, GreedyBytes)

CURVE_SECP256K1 = 1
CURVE_PRIME256R1 = 2
CURVE_ED25519 = 4
CURVE_SLIP21 = (
    8  # not really a curve, but used to indicate the presence of SLIP-21 paths
)
CURVE_BLS12381G1 = 16

Curve = FlagsEnum(
    Byte,
    secp256k1=CURVE_SECP256K1,
    prime256r1=CURVE_PRIME256R1,
    ed25519=CURVE_ED25519,
    slip21=CURVE_SLIP21,
    bls12381g1=CURVE_BLS12381G1,
)

DerivationPath = Prefixed(
    Asn1Length,
    Struct(
        curve=Curve,
        paths=Optional(GreedyRange(Bip32Path)),
        paths_slip21=Optional(GreedyRange(Slip21Path)),
    ),
)

Dependency = Prefixed(
    Asn1Length, Struct(name=PrefixedString, version=Optional(PrefixedString))
)

Dependencies = Prefixed(Asn1Length, GreedyRange(Dependency))


class BolosTag(enum.IntEnum):
    BOLOS_TAG_APPNAME = 1
    BOLOS_TAG_APPVERSION = 2
    BOLOS_TAG_ICON = 3
    BOLOS_TAG_DERIVEPATH = 4
    BOLOS_TAG_DEPENDENCY = 6


Param = Struct(
    type_=Enum(Byte, BolosTag),
    value=Switch(
        this.type_,
        {
            "BOLOS_TAG_APPNAME": AppName,
            "BOLOS_TAG_APPVERSION": Version,
            "BOLOS_TAG_ICON": Icon,
            "BOLOS_TAG_DERIVEPATH": DerivationPath,
            "BOLOS_TAG_DEPENDENCY": Dependencies,
        },
    ),
)

AppParams = GreedyRange(Param)


def main():
    params1 = AppParams.build(
        [
            {"type_": "BOLOS_TAG_APPNAME", "value": "SSH/PGP Agent"},
            {"type_": "BOLOS_TAG_APPVERSION", "value": "0.0.4"},
            {
                "type_": "BOLOS_TAG_ICON",
                "value": (
                    b"\x01\x00\x00\x00\x00\xFF\xFF\xFF\x00\x00\x18\xFC\x24\x02"
                    b"\x24\x0A\x24\x1A\x7E\x32\x66\x62\x6E\x62\x7E\x32\x00\x1A"
                    b"\x40\x0A\x5F\x02\x5F\x02\x40\x02\x40\xFE\x7F\x00\x00"
                ),
            },
            {
                "type_": "BOLOS_TAG_DERIVEPATH",
                "value": {
                    "curve": Curve.prime256r1 | Curve.ed25519 | Curve.slip21,
                    "paths": ["44'/535348'", "13'", "17'"],
                    "paths_slip21": ["MYPATH"],
                },
            },
        ]
    )

    params2 = (
        b"\x01\x0D\x53\x53\x48\x2F\x50\x47\x50\x20\x41\x67\x65\x6E\x74"
        b"\x02\x05\x30\x2E\x30\x2E\x34\x03\x29\x01\x00\x00\x00\x00\xFF"
        b"\xFF\xFF\x00\x00\x18\xFC\x24\x02\x24\x0A\x24\x1A\x7E\x32\x66"
        b"\x62\x6E\x62\x7E\x32\x00\x1A\x40\x0A\x5F\x02\x5F\x02\x40\x02"
        b"\x40\xFE\x7F\x00\x00\x04\x1c\x0E\x02\x80\x00\x00\x2C\x80\x08"
        b"\x2B\x34\x01\x80\x00\x00\x0D\x01\x80\x00\x00\x11\x87\x00\x4D"
        b"\x59\x50\x41\x54\x48"
    )
    assert params1 == params2


if __name__ == "__main__":
    main()



from contextlib import contextmanager

from .device import Device
from .file import FileDevice
from .hid import HidDevice
from .tcp import TcpDevice

DEVICE_CLASSES = [TcpDevice, HidDevice]

__all__ = [
    "FileDevice",
]


def enumerate_devices():
    devices = []
    for cls in DEVICE_CLASSES:
        devices.extend(cls.enumerate_devices())
    return devices


@contextmanager
def open_device(dev: Device):
    """Open a device in a context manager."""
    try:
        dev.open()
        yield dev
    finally:
        dev.close()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: ledgerwallet/proto/listApps.proto
# Protobuf Python Version: 5.28.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    1,
    '',
    'ledgerwallet/proto/listApps.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n!ledgerwallet/proto/listApps.proto\x12\x08listapps\"F\n\x03\x41pp\x12\r\n\x05\x66lags\x18\x01 \x01(\r\x12\x0c\n\x04hash\x18\x02 \x01(\x0c\x12\x14\n\x0chashCodeData\x18\x03 \x01(\x0c\x12\x0c\n\x04name\x18\x04 \x01(\t\"&\n\x07\x41ppList\x12\x1b\n\x04list\x18\x01 \x03(\x0b\x32\r.listapps.Appb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ledgerwallet.proto.listApps_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_APP']._serialized_start=47
  _globals['_APP']._serialized_end=117
  _globals['_APPLIST']._serialized_start=119
  _globals['_APPLIST']._serialized_end=157
# @@protoc_insertion_point(module_scope)



from abc import ABC, abstractmethod


class LedgerServer(ABC):
    @abstractmethod
    def get_nonce(self) -> bytes:
        pass

    @abstractmethod
    def send_nonce(self, nonce: bytes):
        pass

    @abstractmethod
    def receive_certificate_chain(self):
        pass

    @abstractmethod
    def send_certificate_chain(self, chain):
        pass

    def get_shared_secret(self):
        return None



"""
from dataclasses import dataclass

@dataclass
class HsmScript:
    name: str
    default_params: dict
    use_large_stack: bool = True
"""


class HsmScript(object):
    def __init__(self, name: str, default_params: dict, use_large_stack: bool = True):
        self.name = name
        self.default_params = default_params
        self.use_large_stack = use_large_stack



import logging
from enum import Enum, IntEnum

from construct import (
    Const,
    FlagsEnum,
    Hex,
    IfThenElse,
    Int8ub,
    Int32ub,
    Int32ul,
    Optional,
    PascalString,
    Struct,
    this,
)


class DeviceNames(Enum):
    LEDGER_NANO_S = "Ledger Nano S"
    LEDGER_NANO_X = "Ledger Nano X"
    LEDGER_NANO_SP = "Ledger Nano S+"
    LEDGER_BLUE = "Ledger Blue"
    LEDGER_STAX = "Ledger Stax"
    LEDGER_FLEX = "Ledger Flex"
    LEDGER_APEX_P = "Ledger Apex P"
    LEDGER_APEX_M = "Ledger Apex M"


class LedgerIns(IntEnum):
    SECUINS = 0
    GET_VERSION = 1
    VALIDATE_TARGET_ID = 4
    INITIALIZE_AUTHENTICATION = 0x50
    VALIDATE_CERTIFICATE = 0x51
    GET_CERTIFICATE = 0x52
    MUTUAL_AUTHENTICATE = 0x53
    ONBOARD = 0xD0
    RUN_APP = 0xD8
    # Commands for custom endorsement
    ENDORSE_SET_START = 0xC0
    ENDORSE_SET_COMMIT = 0xC2


class LedgerSecureIns(IntEnum):
    SET_LOAD_OFFSET = 5
    LOAD = 6
    FLUSH = 7
    CRC = 8
    COMMIT = 9
    CREATE_APP = 11
    DELETE_APP = 12
    LIST_APPS = 14
    LIST_APPS_CONTINUE = 15
    GET_VERSION = 16
    GET_MEMORY_INFORMATION = 17
    SETUP_CUSTOM_CERTIFICATE = 18
    RESET_CUSTOM_CERTIFICATE = 19
    DELETE_APP_BY_HASH = 21
    MCU_BOOTLOADER = 0xB0


VersionInfo = Struct(
    target_id=Hex(Int32ub),
    se_version=PascalString(Int8ub, "utf-8"),
    _flags_len=Const(b"\x04"),
    flags=FlagsEnum(
        Int32ul,
        recovery_mode=1,
        signed_mcu=2,
        is_onboarded=4,
        trust_issuer=8,
        trust_custom_ca=16,
        hsm_initialized=32,
        pin_validated=128,
    ),
    mcu_version=PascalString(Int8ub, "utf-8"),
    mcu_bl_version=Optional(PascalString(Int8ub, "utf-8")),
    hw_version=Optional(
        IfThenElse(
            this.target_id == 0x33000004, PascalString(Int8ub, "utf-8"), Const(b"")
        )
    ),
    language=Optional(PascalString(Int8ub, "utf-8")),
    _recover_state_len=Optional(Hex(Int8ub)),
    recover_state=Optional(Hex(Int8ub)),
)


def enable_apdu_log():
    logger = logging.getLogger("ledgerwallet")
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging.StreamHandler())


def serialize(buffer: bytes):
    return bytes([len(buffer)]) + buffer


def unserialize(buffer: bytes):
    buffer_len = buffer[0]
    assert len(buffer) >= buffer_len + 1
    return buffer[1 : buffer_len + 1], buffer[buffer_len + 1 :]


def get_device_name(target_id: int) -> str:
    target_ids = {
        0x31100002: DeviceNames.LEDGER_NANO_S.value,  # firmware version <= 1.3.1
        0x31100003: DeviceNames.LEDGER_NANO_S.value,  # firmware version > 1.3.1
        0x31100004: DeviceNames.LEDGER_NANO_S.value,  # firmware version >= 1.5
        0x31000002: DeviceNames.LEDGER_BLUE.value,  # firmware version <= 2.0
        0x31010004: DeviceNames.LEDGER_BLUE.value,  # firmware version > 2.0
        0x33000004: DeviceNames.LEDGER_NANO_X.value,
        0x33100004: DeviceNames.LEDGER_NANO_SP.value,
        0x33200004: DeviceNames.LEDGER_STAX.value,
        0x33300004: DeviceNames.LEDGER_FLEX.value,
        0x33400004: DeviceNames.LEDGER_APEX_P.value,
        0x33500004: DeviceNames.LEDGER_APEX_M.value,
    }
    return target_ids.get(target_id, "Unknown device")


def decode_flags(flags: int):
    flag_values = {
        1: "issuer",
        2: "bolos_upgrade",
        4: "signed",
        8: "bolos_ux",
        16: "derive_master",
        64: "global_pin",
        128: "debug",
        256: "autoboot",
        512: "bolos_settings",
        1024: "custom_ca",
        2048: "library",
        4096: "no_run",
        # "enabled" is always set. no need to display it
        # 32768: "enabled"
    }
    enabled_flags = []
    for f in flag_values:
        if flags & f == f:
            enabled_flags.append(flag_values[f])
    return enabled_flags


def flags_to_string(flags: int):
    return ",".join(decode_flags(flags))
